@@ -1,3 +1,18 @@
+# Copyright 2022 The GPJax Contributors. All Rights Reserved.
+#
+# Licensed under the Apache License, Version 2.0 (the "License");
+# you may not use this file except in compliance with the License.
+# You may obtain a copy of the License at
+#
+#     http://www.apache.org/licenses/LICENSE-2.0
+#
+# Unless required by applicable law or agreed to in writing, software
+# distributed under the License is distributed on an "AS IS" BASIS,
+# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+# See the License for the specific language governing permissions and
+# limitations under the License.
+# ==============================================================================
+
 import abc
 from typing import Callable, Dict
 
@@ -7,13 +22,14 @@ from chex import dataclass
 from jax import vmap
 from jaxtyping import Array, Float
 
+from jaxlinop import identity
+
+from .config import get_defaults
 from .gps import AbstractPosterior
-from .kernels import cross_covariance, diagonal, gram
 from .likelihoods import Gaussian
-from .parameters import transform
 from .quadrature import gauss_hermite_quadrature
-from .types import Dataset
-from .utils import I, concat_dictionaries
+from .types import Dataset, PRNGKeyType
+from .utils import concat_dictionaries
 from .variational_families import (
     AbstractVariationalFamily,
     CollapsedVariationalGaussian,
@@ -31,7 +47,7 @@ class AbstractVariationalInference:
         self.prior = self.posterior.prior
         self.likelihood = self.posterior.likelihood
 
-    def _initialise_params(self, key: jnp.DeviceArray) -> Dict:
+    def _initialise_params(self, key: PRNGKeyType) -> Dict:
         """Construct the parameter set used within the variational scheme adopted."""
         hyperparams = concat_dictionaries(
             {"likelihood": self.posterior.likelihood._initialise_params(key)},
@@ -41,13 +57,13 @@ class AbstractVariationalInference:
 
     @abc.abstractmethod
     def elbo(
-        self, train_data: Dataset, transformations: Dict
+        self,
+        train_data: Dataset,
     ) -> Callable[[Dict], Float[Array, "1"]]:
         """Placeholder method for computing the evidence lower bound function (ELBO), given a training dataset and a set of transformations that map each parameter onto the entire real line.
 
         Args:
             train_data (Dataset): The training dataset for which the ELBO is to be computed.
-            transformations (Dict): A set of functions that unconstrain each parameter.
 
         Returns:
             Callable[[Array], Array]: A function that computes the ELBO given a set of parameters.
@@ -65,27 +81,26 @@ class StochasticVI(AbstractVariationalInference):
         self.num_inducing = self.variational_family.num_inducing
 
     def elbo(
-        self, train_data: Dataset, transformations: Dict, negative: bool = False
+        self, train_data: Dataset, negative: bool = False
     ) -> Callable[[Float[Array, "N D"]], Float[Array, "1"]]:
         """Compute the evidence lower bound under this model. In short, this requires evaluating the expectation of the model's log-likelihood under the variational approximation. To this, we sum the KL divergence from the variational posterior to the prior. When batching occurs, the result is scaled by the batch size relative to the full dataset size.
 
         Args:
             train_data (Dataset): The training data for which we should maximise the ELBO with respect to.
-            transformations (Dict): The transformation set that unconstrains each parameter.
             negative (bool, optional): Whether or not the resultant elbo function should be negative. For gradient descent where we minimise our objective function this argument should be true as minimisation of the negative corresponds to maximisation of the ELBO. Defaults to False.
 
         Returns:
             Callable[[Dict, Dataset], Array]: A callable function that accepts a current parameter estimate and batch of data for which gradients should be computed.
         """
+
+        # Constant for whether or not to negate the elbo for optimisation purposes
         constant = jnp.array(-1.0) if negative else jnp.array(1.0)
 
         def elbo_fn(params: Dict, batch: Dataset) -> Float[Array, "1"]:
-            params = transform(params, transformations)
-
             # KL[q(f(·)) || p(f(·))]
             kl = self.variational_family.prior_kl(params)
 
-            # ∫[log(p(y|f(x))) q(f(x))] df(x)
+            # ∫[log(p(y|f(·))) q(f(·))] df(·)
             var_exp = self.variational_expectation(params, batch)
 
             # For batch size b, we compute  n/b * Σᵢ[ ∫log(p(y|f(xᵢ))) q(f(xᵢ)) df(xᵢ)] - KL[q(f(·)) || p(f(·))]
@@ -105,19 +120,23 @@ class StochasticVI(AbstractVariationalInference):
         Returns:
             Array: The expectation of the model's log-likelihood under our variational distribution.
         """
+
+        # Unpack training batch
         x, y = batch.X, batch.y
 
-        # q(f(x))
-        predictive_dist = vmap(self.variational_family.predict(params))(x[:, None])
-        mean = predictive_dist.mean().val.reshape(-1, 1)
-        variance = predictive_dist.variance().val.reshape(-1, 1)
+        # Variational distribution q(f(·)) = N(f(·); μ(·), Σ(·, ·))
+        q = self.variational_family(params)
+
+        # Compute variational mean, μ(x), and variance, √diag(Σ(x, x)), at training inputs, x
+        def q_moments(x):
+            qx = q(x)
+            return qx.mean(), qx.variance()
+
+        mean, variance = vmap(q_moments)(x[:, None])
 
         # log(p(y|f(x)))
-        log_prob = vmap(
-            lambda f, y: self.likelihood.link_function(
-                f, params["likelihood"]
-            ).log_prob(y)
-        )
+        link_function = self.likelihood.link_function
+        log_prob = vmap(lambda f, y: link_function(params["likelihood"], f).log_prob(y))
 
         # ≈ ∫[log(p(y|f(x))) q(f(x))] df(x)
         expectation = gauss_hermite_quadrature(log_prob, mean, variance, y=y)
@@ -142,37 +161,44 @@ class CollapsedVI(AbstractVariationalInference):
             raise TypeError("Variational family must be CollapsedVariationalGaussian.")
 
     def elbo(
-        self, train_data: Dataset, transformations: Dict, negative: bool = False
-    ) -> Callable[[dict], Float[Array, "1"]]:
+        self, train_data: Dataset, negative: bool = False
+    ) -> Callable[[Dict], Float[Array, "1"]]:
         """Compute the evidence lower bound under this model. In short, this requires evaluating the expectation of the model's log-likelihood under the variational approximation. To this, we sum the KL divergence from the variational posterior to the prior. When batching occurs, the result is scaled by the batch size relative to the full dataset size.
 
         Args:
             train_data (Dataset): The training data for which we should maximise the ELBO with respect to.
-            transformations (Dict): The transformation set that unconstrains each parameter.
             negative (bool, optional): Whether or not the resultant elbo function should be negative. For gradient descent where we minimise our objective function this argument should be true as minimisation of the negative corresponds to maximisation of the ELBO. Defaults to False.
 
         Returns:
             Callable[[Dict, Dataset], Array]: A callable function that accepts a current parameter estimate for which gradients should be computed.
         """
-        constant = jnp.array(-1.0) if negative else jnp.array(1.0)
 
+        # Unpack training data
         x, y, n = train_data.X, train_data.y, train_data.n
 
+        # Unpack mean function and kernel
+        mean_function = self.prior.mean_function
+        kernel = self.prior.kernel
+
+        # Unpack kernel computation
+        gram, cross_covariance = kernel.gram, kernel.cross_covariance
+
         m = self.num_inducing
+        jitter = get_defaults()["jitter"]
+
+        # Constant for whether or not to negate the elbo for optimisation purposes
+        constant = jnp.array(-1.0) if negative else jnp.array(1.0)
 
         def elbo_fn(params: Dict) -> Float[Array, "1"]:
-            params = transform(params, transformations)
             noise = params["likelihood"]["obs_noise"]
             z = params["variational_family"]["inducing_inputs"]
-            Kzz = gram(self.prior.kernel, z, params["kernel"])
-            Kzz += I(m) * self.variational_family.jitter
-            Kzx = cross_covariance(self.prior.kernel, z, x, params["kernel"])
-            Kxx_diag = vmap(self.prior.kernel, in_axes=(0, 0, None))(
-                x, x, params["kernel"]
-            )
-            μx = self.prior.mean_function(x, params["mean_function"])
+            Kzz = gram(kernel, params["kernel"], z)
+            Kzz += identity(m) * jitter
+            Kzx = cross_covariance(kernel, params["kernel"], z, x)
+            Kxx_diag = vmap(kernel, in_axes=(None, 0, 0))(params["kernel"], x, x)
+            μx = mean_function(params["mean_function"], x)
 
-            Lz = jnp.linalg.cholesky(Kzz)
+            Lz = Kzz.to_root()
 
             # Notation and derivation:
             #
@@ -198,13 +224,13 @@ class CollapsedVI(AbstractVariationalInference):
             #
             #   with A and B defined as above.
 
-            A = jsp.linalg.solve_triangular(Lz, Kzx, lower=True) / jnp.sqrt(noise)
+            A = Lz.solve(Kzx) / jnp.sqrt(noise)
 
             # AAᵀ
             AAT = jnp.matmul(A, A.T)
 
             # B = I + AAᵀ
-            B = I(m) + AAT
+            B = jnp.eye(m) + AAT
 
             # LLᵀ = I + AAᵀ
             L = jnp.linalg.cholesky(B)
@@ -232,3 +258,10 @@ class CollapsedVI(AbstractVariationalInference):
             return constant * (two_log_prob - two_trace).squeeze() / 2.0
 
         return elbo_fn
+
+
+__all__ = [
+    "AbstractVariationalInference",
+    "StochasticVI",
+    "CollapsedVI",
+]
